<?php
/*
 * ZX Spectrum SCREEN converter to PNG or animated GIF file
 * Copyright (C) 2008-2009 Andrew nyuk Marinov (aka.nyuk@gmail.com)
 *
 * v0.4 (2009-10-04) 
 *
 * Some code parts taken from:
 * - theX.pvd decoder (http://thex.untergrund.net/pvd/)
 * - moroz1999 sources (http://www.zx.pk.ru/member.php?u=50)
 */

class ZXGFX {
	private $palettes = array(
		'pulsar' 	=> 	'00,76,CD,E9,FF,9F:FF,00,00;00,FF,00;00,00,FF',
		'alone' 	=>	'00,60,A0,E0,FF,A0:FF,00,00;00,FF,00;00,00,FF',
		'orthodox' 	=> 	'00,76,CD,E9,FF,9F:D0,00,00;00,E4,00;00,00,FF',
		'Mars'		=> 	'00,80,C0,E0,FF,C8:FF,00,00;40,C0,00;00,40,C0',
		'Ocean'		=>	'20,80,A0,C0,E0,A8:D0,00,30;00,D0,30;00,00,FF',
		'Grey'		=>	'00,80,C0,E0,FF,C8:49,92,24;49,92,24;49,92,24'	
	);
	
	private $borders = array(
		'none' => array(
			'0.5' => array('width' => 128, 'height' => 96),
			'1' => array('width' => 256, 'height' => 192),
			'2' => array('width' => 512, 'height' => 384),
			'3' => array('width' => 768, 'height' => 576),
		),
		'small' => array(
			'0.5' => array('width' => 160, 'height' => 120),
			'1' => array('width' => 320, 'height' => 240),
			'2' => array('width' => 640, 'height' => 480),
			'3' => array('width' => 960, 'height' => 720),
			
		),
		'complete' => array(
			'0.5' => array('width' => 192, 'height' => 150),
			'1' => array('width' => 384, 'height' => 300),
			'2' => array('width' => 768, 'height' => 600),
			'3' => array('width' => 1152, 'height' => 900),
		),
	);

	// Border size
	private $border = 'none';
	
	private $border_color = 0;
	
	// Generated from palette colors
	private $colors = false;
	
	// Generated from palette gigascreen colors
	private $gigaColors = false;
	
	private $mask = array(0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01);
	
	// Исходный битмап
	private $src;
	
	private $input_type = 'normal';		// Image input format
	private $MGS_CharSize = 1;				// MGS screen format
	private $is_flash_img	= false;	// Is loaded image with FLASH attribute
	private $output_type = 'png';
	private $output_scale = 1;

	// FLASH attribute -> GIF animation related
	private $animation_delay = 50;
	
	private $options = array(
		'forceOneFrame' => false,		// Force remove secon frame in GIF-images
		'showHiddenPixels' => false,	// Show hidden pixels (INK=PAPER) as transparent
	);
	
	private function generateColors($palette, $S = 0x100) {
		for ($i=0; $i<16; $i++) {
			$one = ($i & 0x08) ? $palette['BB'] : $palette['NN'];

			$r = ($i & 0x02) ? $one : $palette['ZZ'];
			$g = ($i & 0x04) ? $one : $palette['ZZ'];
			$b = ($i & 0x01) ? $one : $palette['ZZ'];
			
			$redChannel = round(($r*$palette['R11'] + $g*$palette['R12'] + $b*$palette['R13'])/$S);
			$greenChannel = intval(($r*$palette['R21'] + $g*$palette['R22'] + $b*$palette['R23'])/$S);
			$blueChannel = intval(($r*$palette['R31'] + $g*$palette['R32'] + $b*$palette['R33'])/$S);

			$this->colors[$i] = $redChannel*0x010000 + $greenChannel*0x0100 + $blueChannel;
		}
	}
	
	private function generateGigaColors($palette, $S = 0x100) {
		$cache = array(0 => 'Z', 1 => 'N', 2 => 'Z', 3 => 'B');

		$palette['BN'] = $palette['NB'];
		$palette['BZ'] = $palette['ZB'];
		$palette['NZ'] = $palette['ZN'];
		
		for ($i1=0; $i1<16; $i1++) for ($i2=0; $i2<16; $i2++) {
			$brightness1 = ($i1 & 0x08) ? 1 : 0;
			$brightness2 = ($i2 & 0x08) ? 1 : 0;			
			
			$r = $palette[$cache[($brightness1 * 2) + (($i1 & 0x02) ? 1 : 0)].$cache[($brightness2 * 2) + (($i2 & 0x02) ? 1 : 0)]];
			$g = $palette[$cache[($brightness1 * 2) + (($i1 & 0x04) ? 1 : 0)].$cache[($brightness2 * 2) + (($i2 & 0x04) ? 1 : 0)]];
			$b = $palette[$cache[($brightness1 * 2) + (($i1 & 0x01) ? 1 : 0)].$cache[($brightness2 * 2) + (($i2 & 0x01) ? 1 : 0)]];
			
			$redChannel = round(($r*$palette['R11'] + $g*$palette['R12'] + $b*$palette['R13'])/$S);
			$greenChannel = round(($r*$palette['R21'] + $g*$palette['R22'] + $b*$palette['R23'])/$S);
			$blueChannel = round(($r*$palette['R31'] + $g*$palette['R32'] + $b*$palette['R33'])/$S);

			$this->gigaColors[$i1][$i2] = $redChannel*0x010000 + $greenChannel*0x0100 + $blueChannel;
		}
	}
	
	private function detectInputType() {
		if (strlen($this->src) == 36871 && substr($this->src, 0, 3) == 'MGS') {
			$this->input_type = 'mgs';			
			$this->MGS_CharSize	= ord($this->src[4]);
			if (!in_array($this->MGS_CharSize, array(1,2,4,8))) return false;
		}
		elseif (strlen($this->src) == 13824) {
			$this->input_type = 'gigascreen';			
		}
		elseif (strlen($this->src) == 6912) { 
			$this->input_type = 'screen';
		}
		elseif (strlen($this->src) == 6144) {
			$this->input_type = 'bw-screen';			
		}
		else
			return false;
		
		return $this->input_type;
	}
	
	private function generateImage() {
		// Save values
		$save_colors = $this->colors;
		$save_border = $this->border_color;

		if (!$this->colors) $this->setPalette();
		
		$options = array();
		$frame2 = false;

		// Multigigascreen
		if ($this->input_type == 'mgs') {
			$frame = $this->createMgsFrame($options);
		}
		// Gigascreen
		elseif ($this->input_type == 'gigascreen') {
			$frame = $this->createGigascreenFrame($options);
		}
		// Normal screen or screen without attributes
		else {
			$frame = $this->createScreenFrame($options);
			if ($this->output_type == 'gif' && $this->is_flash_img) {
				$options['secondFrame'] = 1;
				$frame2 = $this->createScreenFrame($options);
			}
		}

	    if ($this->output_type == 'gif') {
			// First frame
			ob_start();
            imagegif ($frame);
            $gif_frames[] = ob_get_contents();
            ob_end_clean();
            $this->destroyFrame($frame);

            if ($frame2 && !$this->options['forceOneFrame']) {
				// Second frame
				ob_start();
	            imagegif ($frame2);
	            $gif_frames[] = ob_get_contents();
	            ob_end_clean();
				if ($frame2) $this->destroyFrame($frame2);
					            
		        $GIF = new GIFEncoder($gif_frames, array($this->animation_delay, $this->animation_delay), 0, 2, -1, -1, -1, "bin");
		        $result_img = $GIF->GetAnimation();
            }
            else {
            	$result_img = $gif_frames[0];
            }
        }
        // 'png' by default
		else {
	    	ob_start();
			imagepng($frame);
			$result_img = ob_get_contents();
        	ob_end_clean();
        	$this->destroyFrame($frame);
	    }		
		
	    // Restore values
	    $this->colors = $save_colors;
	    $this->border_color = $save_border;
	    
        return $result_img;		
	}
	
	private function destroyFrame($im) {
		imagedestroy($im);
	}

	private function createScreenFrame($options = array()) {
		// By default we process only one frame
		$this->is_flash_img = false;
		
		$im = @imagecreatetruecolor(256, 192) or die("Cannot Initialize new GD image stream");
		imagefill($im, 0, 0, 0);
		
		if ($this->options['showHiddenPixels']) {
			$transparent_color = imagecolorallocate($im, 1,1,1);
			imagecolortransparent($im, $transparent_color);
		}
		
		for($y = 0; $y < 192; ++$y) {
			$pixelLine = 32 * (($y & 0xc0) | (($y << 3) & 0x38) | (($y >> 3) & 0x07));
	      	$attr      = 6144 + (($y & 0xf8) << 2);
	
	      	for($x = 0; $x < 32; ++$x) {
	        	$chr_pixels   = ord($this->src[$pixelLine + $x]);
	        	$chr_attr     = ord($this->src[$attr + $x]);
	        
	        	if ($this->input_type == 'bw-screen') {
		        	$paper = 7;
		        	$ink   = 0;
	        	}
	        	else {
		        	$paper = ($chr_attr >> 3) & 0x0f;
		        	$ink   = ($chr_attr & 0x07) | ($paper & 0x08);
	        	}
	        	
	        	if (($chr_attr & 0x80) && $paper != $ink) $this->is_flash_img = true;
					
				for ($bit = 0; $bit < 8; ++$bit) {
					if ($this->options['showHiddenPixels'] && $ink == $paper && ($chr_pixels & $this->mask[$bit])) {
						$color = $transparent_color;
					}
					// Flashing 		          
		          	elseif ($options['secondFrame'] && ($chr_attr & 0x80)) {
		            	$color = ($chr_pixels & $this->mask[$bit]) ? $this->colors[$paper] : $this->colors[$ink];
		          	}
		          	else {
		          		$color = ($chr_pixels & $this->mask[$bit]) ? $this->colors[$ink] : $this->colors[$paper];
		          	}

		        	if ($color) {
						imagesetpixel($im, $x*8 + $bit, $y, $color);
		        	}
	        	}
	    	}
		}
	    
	    return $this->finalizeFrame($im);;
	}
	
	
	private function createGigascreenFrame($options = array()) {
		$im = @imagecreatetruecolor(256, 192) or die("Cannot Initialize new GD image stream");
		imagefill($im, 0, 0, 0);
		if ($this->options['showHiddenPixels']) {
			$transparent_color = imagecolorallocate($im, 1,1,1);
			imagecolortransparent($im, $transparent_color);
		}
		
		for($y = 0; $y < 192; ++$y) {
			$pixelLine = 32 * (($y & 0xc0) | (($y << 3) & 0x38) | (($y >> 3) & 0x07));
	      	$attr      = 6144 + (($y & 0xf8) << 2);
	
	      	for($x = 0; $x < 32; ++$x) {
	        	$chr_pixels1   = ord($this->src[$pixelLine + $x]);
	        	$paper1 = (ord($this->src[$attr + $x]) >> 3) & 0x0f;
	        	$ink1   = (ord($this->src[$attr + $x]) & 0x07) | ($paper1 & 0x08);

	        	$chr_pixels2   = ord($this->src[$pixelLine + $x + 6912]);
	        	$paper2 = (ord($this->src[$attr + $x + 6912]) >> 3) & 0x0f;
	        	$ink2   = (ord($this->src[$attr + $x + 6912]) & 0x07) | ($paper2 & 0x08);
	        	
				for($bit = 0; $bit < 8; ++$bit) {
		        	$ci1 = ($chr_pixels1 & $this->mask[$bit]) ? $ink1 : $paper1;
		        	$ci2 = ($chr_pixels2 & $this->mask[$bit]) ? $ink2 : $paper2;
		        	
		        	$trans1 = ($this->options['showHiddenPixels'] && $ink1 == $paper1 && ($chr_pixels1 & $this->mask[$bit])) ? 1 : 0;
		        	$trans2 = ($this->options['showHiddenPixels'] && $ink2 == $paper2 && ($chr_pixels1 & $this->mask[$bit])) ? 1 : 0;
		        	
					if ($trans1 || $trans2) {
						imagesetpixel($im, $x*8 + $bit, $y, $transparent_color);
					}
					elseif ($this->gigaColors[$ci1][$ci2]) {
						imagesetpixel($im, $x*8 + $bit, $y, $this->gigaColors[$ci1][$ci2]);
		        	}
	        	}
	    	}
		}
	    
	    return $this->finalizeFrame($im);
	}

	private function createMgsFrame($options = array()) {
		$im = @imagecreatetruecolor(256, 192) or die("Cannot Initialize new GD image stream");
		imagefill($im, 0, 0, 0);
		if ($this->options['showHiddenPixels']) {
			$transparent_color = imagecolorallocate($im, 1,1,1);
			imagecolortransparent($im, $transparent_color);
		}
		
		// mgs colors => $this->gigaScreen
		$colors = array(0 => 0,	1 => 1,	2 => 9,	3 => 2,	4 => 10, 5 => 3, 6 => 11, 7 => 4, 8 => 12, 9 => 5, 10 => 13, 11 => 6, 12 => 14,	13 => 7, 14 => 15);
		
		// Set border color
		$b1 = ord($this->src[5]);
		$b2 = ord($this->src[6]);
		if (isset($this->gigaColors[$b1][$b2])) {
			$this->border_color = $this->gigaColors[$b1][$b2];
		} 
		
		for($y = 0; $y < 192; ++$y) {
			$pixelLine 	= $y * 32 + 7;
	      	$attr		= 12295 + $y * 64 + ($y & 0x07 * 1536);
			
	      	for($x = 0; $x < 32; ++$x) {
	        	$chr_pixels1   = ord($this->src[$pixelLine + $x]);
	        	$chr_pixels2   = ord($this->src[$pixelLine + $x + 6144]);
	        	
	        	switch ($this->MGS_CharSize) {
	        		case 1:
		        		if ($x < 8 || $x >= 23) {
				        	$ink1   = $colors[ord($this->src[12295 + ($y >> 3)*512 + $x*2])];
				        	$paper1 = $colors[ord($this->src[12296 + ($y >> 3)*512 + $x*2])];
				        	$ink2   = $colors[ord($this->src[24583 + ($y >> 3)*512 + $x*2])];
				        	$paper2 = $colors[ord($this->src[24584 + ($y >> 3)*512 + $x*2])];
		        		}
			        	else { 
				        	$ink1   = $colors[ord($this->src[12295 + $y*64 + $x*2])];
				        	$paper1 = $colors[ord($this->src[12296 + $y*64 + $x*2])];
				        	$ink2   = $colors[ord($this->src[24583 + $y*64 + $x*2])];
				        	$paper2 = $colors[ord($this->src[24584 + $y*64 + $x*2])];
			        	}
			        	break;
	        		case 2:
			        	$ink1   = $colors[ord($this->src[12295 + ($y >> 1)*128 + $x*2])];
			        	$paper1 = $colors[ord($this->src[12296 + ($y >> 1)*128 + $x*2])];
			        	$ink2   = $colors[ord($this->src[24583 + ($y >> 1)*128 + $x*2])];
			        	$paper2 = $colors[ord($this->src[24584 + ($y >> 1)*128 + $x*2])];
			        	break;
	        		case 4:
			        	$ink1   = $colors[ord($this->src[12295 + ($y >> 2)*256 + $x*2])];
			        	$paper1 = $colors[ord($this->src[12296 + ($y >> 2)*256 + $x*2])];
			        	$ink2   = $colors[ord($this->src[24583 + ($y >> 2)*256 + $x*2])];
			        	$paper2 = $colors[ord($this->src[24584 + ($y >> 2)*256 + $x*2])];
			        	break;
	        		case 8:
			        	$ink1   = $colors[ord($this->src[12295 + ($y >> 3)*512 + $x*2])];
			        	$paper1 = $colors[ord($this->src[12296 + ($y >> 3)*512 + $x*2])];
			        	$ink2   = $colors[ord($this->src[24583 + ($y >> 3)*512 + $x*2])];
			        	$paper2 = $colors[ord($this->src[24584 + ($y >> 3)*512 + $x*2])];
			        	break;
	        	}
	        	
	        	
	        	for($bit = 0; $bit < 8; ++$bit) {
		        	$ci1 = ($chr_pixels1 & $this->mask[$bit]) ? $ink1 : $paper1;
		        	$ci2 = ($chr_pixels2 & $this->mask[$bit]) ? $ink2 : $paper2;
		        	
		        	if ($this->options['showHiddenPixels'] && ($ink1 == $paper1 && ($chr_pixels1 & $this->mask[$bit]) || ($ink2 == $paper2 && ($chr_pixels1 & $this->mask[$bit])))) {
		        		$color = $transparent_color;
		        	}
		        	elseif($this->gigaColors[$ci1][$ci2]) {
		        		$color = $this->gigaColors[$ci1][$ci2];
		        	}
		        	else
		        		$color = false;
		        		
		        	if ($color)	imagesetpixel($im, $x*8 + $bit, $y, $color);
	        	}
	    	}
		}
	    
	    return $this->finalizeFrame($im);
	}

	/**
	 * Resize generated image and add border
	 * @param $frame
	 * @return unknown_type
	 */
	private function finalizeFrame($imtmp) {
		$full_width = $this->borders[$this->border][strval($this->output_scale)]['width'];
		$full_height = $this->borders[$this->border][strval($this->output_scale)]['height'];
		
		$image_left = intval(($full_width - $this->output_scale * 256)/2);
		$image_top = intval(($full_height - $this->output_scale * 192)/2);
		
		$frame = @imagecreatetruecolor($full_width, $full_height) or die("Cannot Initialize new GD image stream1");
		
		if ($this->options['showHiddenPixels']) {
			$transparent_color = imagecolorallocate($frame, 1,1,1);
			imagecolortransparent($frame, $transparent_color);
			imagefill($frame, 0, 0, $transparent_color);
		}
		else {
			imagefill($frame, 0, 0, $this->border_color);
		}
		
		if ($this->output_scale != 1) {
			imagecopyresampled ($frame, $imtmp, $image_left, $image_top + $y * $this->output_scale, 0, 0, $this->output_scale * 256, $this->output_scale * 192, 256, 192);
		}
		else {
			imagecopy ($frame, $imtmp, $image_left, $image_top + $y, 0, 0, 256, 192);
		}
		
		imagedestroy($imtmp);
		return $frame;
	}
		
	/**
	 * Return all available borders sizes
	 * 
	 * @return array
	 */
	function getBorders() {
		return $this->borders;
	}

	/**
	 * Set border size
	 *  
	 * @param $str string is none, small, media
	 * @return void
	 */
	function setBorder($str = '') {
		if (isset($this->borders[$str])) { 
			$this->border = $str;
		}
	}
	
	/**
	 * Set border color
	 * 
	 * @param $color integer from 0 to 7
	 * @return void
	 */
	function setBorderColor($color = 0) {
		if ($color >= 0 && $color <= 7) {
			$this->border_color = $this->colors[intval($color)];
		}
	}
	
	/**
	 * Return all available palettes
	 * 
	 * @return array
	 */
	function getPalettes() {
		return $this->palettes;
	}
	
	/**
	 * Set given palette as work palette
	 * 
	 * @param $str string palette name
	 * @return void
	 */
	function setPalette($str = '') {
		if ($this->input_type == 'bw-screen') {
			$str_palette = $this->palettes['Grey'];
		}
		elseif (isset($this->palettes[$str])) { 
			$str_palette = $this->palettes[$str];
		}
		elseif (!$str) {
			$str_palette = reset($this->palettes);
		}
		else {
			$str_palette = $str;
		}
		
		list($bright, $colors) = split(':', $str_palette);
		
		list($red, $green, $blue) = split(';', $colors);
		list($palette['ZZ'], $palette['ZN'], $palette['NN'], $palette['NB'], $palette['BB'], $palette['ZB']) = split(',', $bright);
		
		list($palette['R11'], $palette['R12'], $palette['R13']) = split(',', $red);
		list($palette['R21'], $palette['R22'], $palette['R23']) = split(',', $green);
		list($palette['R31'], $palette['R32'], $palette['R33']) = split(',', $blue);
		
		foreach($palette as &$val) {
			$val = hexdec(trim($val));			
		}
		
		$Sr = $palette['R11']+$palette['R12']+$palette['R13'];
		$Sg = $palette['R21']+$palette['R22']+$palette['R23'];
		$Sb = $palette['R31']+$palette['R32']+$palette['R33'];
		$S = max($Sr, $Sg, $Sb);
		
		$this->generateGigaColors($palette, $S);
		$this->generateColors($palette, $S);
	}

	/**
	 * Return output type
	 * 
	 * @return string with 'png' or 'gif'
	 */
	function getOutputType() {
		return $this->output_type;
	}

	
	/**
	 * Force output type
	 * 
	 * @param $type string with 'png' or 'gif'
	 * @return void
	 */
	function setOutputType($type) {
		$type = strtolower($type);
		if ($type == 'png' || $type == 'gif') 
			$this->output_type = $type;
	}

	/**
	 * Set output image scale
	 * 
	 * @param $scale integer of 1, 2, 3
	 * @return void
	 */
	function setOutputScale($scale = 1) {
		if (in_array($scale, array(0.5,2,3))) {
			$this->output_scale = floatval($scale);
		}
	}

	/**
	 * Set option
	 * 
	 * @param $option_name
	 * @param $value
	 * @return void
	 */
	function setOption($option_name, $value = false) {
		if (isset($this->options[$option_name])) {
			$this->options[$option_name] = $value;
		}
	}
	
	/**
	 * Load source image
	 * 
	 * @param $data	string with data
	 * @return string with type of image. If type not detected, return  false
	 */
	function loadData($data) {
		if (!$data) return false;
		$this->src = $data;
		 
		return ($this->detectInputType()) ? $this->input_type : false;
	}

	/**
	 * 	Generate image and return it
	 * 
	 */
	function generate() {
		return $this->generateImage();
	}
	
	/**
	 * 	Generate image and show
	 * 
	 */
	function show() {
		$result = $this->generateImage();

		header("Pragma: no-cache");
	    if ($this->output_type == 'png') {
		    header("Content-type: image/png");
		    echo $result;
			exit;
	    }		
	    if ($this->output_type == 'gif') {
	        header ('Content-type: image/gif');
		    echo $result;
			exit;
        }
	}

	/**
	 * 	Generate image and force download
	 * 
	 */
	function download($filename = 'converted') {
		$result = $this->generateImage();

	    if ($this->output_type == 'png') {
		    header("Content-type: image/png");
			header('Content-Disposition: attachment; filename="'.$filename.'.png"');
			header("Content-Transfer-Encoding: binary");
			header("Content-Length: ".strlen($result));
			
		    echo $result;
			exit;
	    }		
	    if ($this->output_type == 'gif') {
	        header ('Content-type: image/gif');
			header('Content-Disposition: attachment; filename="'.$filename.'.gif"');
			header("Content-Transfer-Encoding: binary");
			header("Content-Length: ".strlen($result));
			echo $result;
			exit;
        }
	}

	/**
	 * 	Generate image and save it on server
	 * 
	 */
	function save($path) {
		$result = $this->generateImage();
		
		$path .= '.'.$this->output_type;
    	if (!$handle = fopen($path, 'w')) return false; 

    	if (fwrite($handle, $result) === FALSE)	return false;
    	fclose($handle);
    	
    	return $path;
	}	
}	
?>